import argparse
import logging
from importlib.metadata import version
from pathlib import Path

from dvtag import get_rjid, tag
from utils import wav_to_flac, wav_to_mp3

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s", datefmt="%Y-%m-%d %a %H:%M:%S"
)


def start(dirpath: Path, w2f: bool, w2m: bool):
    if get_rjid(dirpath.name):
        if w2f:
            wav_to_flac(dirpath)
        if w2m:
            wav_to_mp3(dirpath)
        tag(dirpath)
        return

    for file in dirpath.iterdir():
        if not file.is_dir():
            continue
        start(file, w2f, w2m)


def main():
    parser = argparse.ArgumentParser(prog="dvtag", description="Doujin Voice Tagging Tool (tagging in place)")
    parser.add_argument("-v", "--version", action="version", version=version(parser.prog))
    parser.add_argument("dirpath", type=str, help="a required directory path")
    parser.add_argument(
        "-w2f", default=False, action=argparse.BooleanOptionalAction, help="transcode all wav files to flac [LOSELESS]"
    )
    parser.add_argument(
        "-w2m", default=False, action=argparse.BooleanOptionalAction, help="transcode all wav files to mp3"
    )

    args = parser.parse_args()
    path = Path(args.dirpath).absolute()

    start(path, args.w2f, args.w2m)


if __name__ == "__main__":
    main()



from html import unescape
import json
import logging
import re

from dvtag.utils import create_request_session

session = create_request_session()


class DoujinVoice:
    def __init__(self, rjid: str) -> None:
        self.rjid = rjid

        self.dl_count = 0
        self.url = ""
        self.work_name = ""
        self.work_image = ""
        self.seiyus = []
        self.circle = ""
        self.genres = []
        self.sale_date = ""

        self._init_metadata()
        self._add_metadata()
        self._get_cover()

    def _add_metadata(self):
        html = session.get(self.url).text

        try:
            pattern = r"<th>声優</th>[\s\S]*?<td>[\s\S]*?(<a[\s\S]*?>[\s\S]*?)</td>"
            seiyu_list_html = re.search(pattern, html).group(1)

            pattern = r"<a[\s\S]*?>(.*?)<"
            for seiyu_html in re.finditer(pattern, seiyu_list_html):
                self.seiyus.append(unescape(seiyu_html.group(1)))
        except AttributeError as e:
            logging.error("Cannot get artists from {}: {}".format(self.rjid, e))

        try:
            pattern = r"<th>サークル名</th>[\s\S]*?<a[\s\S]*?>(.*?)<"
            circle = re.search(pattern, html).group(1)
            self.circle = unescape(circle)

        except AttributeError as e:
            logging.error("Cannot get circle from {}: {}".format(self.rjid, e))

        pattern = r"work\.genre\">(.*)\</a>"
        self.genres = [unescape(m[1]) for m in re.finditer(pattern, html)]

        # get sale date
        pattern = r"www\.dlsite\.com/maniax/new/=/year/([0-9]{4})/mon/([0-9]{2})/day/([0-9]{2})/"
        match = re.search(pattern, html)
        if match:
            self.sale_date = "{}-{}-{}".format(match.group(1), match.group(2), match.group(3))

    def _init_metadata(self):
        rsp = session.get("https://www.dlsite.com/maniax/product/info/ajax?product_id=" + self.rjid)

        try:
            json_data = rsp.json()[self.rjid]

            self.dl_count = int(json_data["dl_count"])
            self.url = json_data["down_url"].replace("download/split", "work").replace("download", "work")
            self.work_name = json_data["work_name"]
            self.work_image = "https:" + json_data["work_image"]

        except ValueError as e:
            logging.error(f"Cannot convert a response to json or convert dl_count to int with RJ-ID {self.rjid}: {e}")
        except KeyError as e:
            logging.error(e)

    def _get_cover(self):
        """
        Tries to fetch a better cover
        """
        try:
            chobit_api = f"https://chobit.cc/api/v1/dlsite/embed?workno={self.rjid}"

            res = json.loads(session.get(chobit_api).text[9:-1])

            self.work_image = res["works"][0]["thumb"].replace("media.dlsite.com/chobit", "file.chobit.cc", 1)

        except Exception as e:
            logging.warning(f"Cannot fetch cover from chobit for {self.rjid}: {e}")



import logging
import os
from pathlib import Path
import subprocess
from typing import List


def transcode_wav(dir: Path, format: str, options: List[str] = []):
    for dirpath, _, filenames in os.walk(dir):
        for filename_wav in filenames:
            if not filename_wav.endswith(".wav"):
                continue
            filename_trans = filename_wav[:-3] + format

            file_wav = os.path.join(dirpath, filename_wav)
            file_trans = os.path.join(dirpath, filename_trans)
            if os.path.exists(file_trans):
                # this should not happen
                logging.warning(f"{filename_trans} already exists.")
                continue

            logging.info(f"Start transcoding {filename_wav} to {format}")

            returncode = subprocess.call(
                ["ffmpeg", "-i", file_wav, *options, file_trans], stdout=open(os.devnull, "w"), stderr=subprocess.STDOUT
            )
            if returncode == 0:
                logging.info(f"Transcoded {filename_wav} successfully, deleting this source file")
                os.remove(file_wav)
            else:
                logging.fatal(f"Failed to transcode {filename_wav} to {format}. Check your ffmpeg")


def wav_to_flac(dir: Path):
    transcode_wav(dir, "flac")


def wav_to_mp3(dir: Path):
    transcode_wav(dir, "mp3", ["-b:a", "320k"])



import io
from io import BytesIO
from pathlib import Path
import re
from typing import List, Optional, Tuple

from PIL import Image
from mutagen.flac import Picture
from mutagen.id3 import PictureType
from natsort import os_sort_key
import requests
from requests.adapters import HTTPAdapter

rjid_pat = re.compile(r"RJ\d{6}(\d\d)?", flags=re.IGNORECASE)


def _split(audio_files: List[Path]) -> List[List[Path]]:
    regexes = [
        r"^omake_?.*[0-9]{1,2}.*$",
        r"^.*ex[0-9]{1,2}.*$",
        r"^ex_.+$",
        r"^後日談.*$",
        r"^おまけ_?[0-9]{0,2}.*$",
        r"^反転おまけ_?[0-9]{1,2}.*$",
        r"^反転_?[0-9]{1,2}.*$",
        r"^20..年?[0-9]{1,2}月配信.*$",
        r"^.*特典.*$",
        r"^追加[0-9]{1,2}.*$",
        r"^opt[0-9]?.*",
        r"^#[0-9]+(-|ー)B",
        r"^#[0-9]+(-|ー)C",
        r"^ASMR_.*",
        r"^.+Bパート",
        r"^番外編",
    ]  # Regular expressions must keep no collision with each other

    results = {}
    paths = []
    regular = []
    for audio_file in audio_files:
        matched = False
        for regex_expr in regexes:
            if re.match(regex_expr, audio_file.stem, re.IGNORECASE):
                matched = True
                if results.get(regex_expr):
                    results[regex_expr].append(audio_file)
                else:
                    results[regex_expr] = [audio_file]
                break

        if not matched:
            regular.append(audio_file)

    if len(regular):
        paths.append(regular)

    for _, v in results.items():
        paths.append(v)

    return paths


def _walk(basepath: Path):
    dirs = []
    files = []
    for file in basepath.iterdir():
        if file.is_dir():
            dirs.append(file)
        else:
            files.append(file)
    yield files

    dirs = sorted(dirs, key=lambda d: os_sort_key(d.name))
    for d in dirs:
        for f in _walk(d):
            yield f


def get_audio_paths_list(basepath: Path) -> Tuple[List[List[Path]], List[List[Path]]]:
    """Gets audio files(Path) from basepath recursively

    Args:
        basepath (Path): base path

    Returns:
        Tuple[List[List[Path]], List[List[Path]]]: flac paths list, mp3 paths list
    """
    flac_paths_list = []
    mp3_paths_list = []

    for files in _walk(basepath):
        mp3_paths = []
        flac_paths = []
        for file in files:
            if file.name.endswith(".flac"):
                flac_paths.append(file)
            elif file.name.endswith(".mp3"):
                mp3_paths.append(file)

        if len(flac_paths):
            flac_paths_list.extend(_split(flac_paths))
        if len(mp3_paths):
            mp3_paths_list.extend(_split(mp3_paths))

    return flac_paths_list, mp3_paths_list


def get_rjid(name: str) -> Optional[str]:
    """Gets rjid(or rather, rj code) from a given string

    Args:
        name (str): A string

    Returns:
        Optional[str]: Returns a string(upper case, like RJ123123) if found, otherwise return None
    """
    m = rjid_pat.search(name)
    if m:
        return m.group().upper()
    return None


def get_image(url: str) -> Image.Image:
    cover_path = create_request_session().get(url, stream=True).raw
    return Image.open(cover_path)


png_modes_to_bpp = {
    "1": 1,
    "L": 8,
    "P": 8,
    "RGB": 24,
    "RGBA": 32,
    "I": 32,
}


def get_png_byte_arr(im: Image.Image) -> BytesIO:
    if im.mode not in png_modes_to_bpp:
        im = im.convert("RGB" if im.info.get("transparency") is None else "RGBA")
    img_byte_arr = io.BytesIO()
    im.save(img_byte_arr, "png")
    return img_byte_arr


def get_picture(png_byte_arr: BytesIO, width: int, height: int, mode: str) -> Picture:
    picture = Picture()
    picture.mime = "image/png"
    picture.width = width
    picture.height = height
    picture.type = PictureType.COVER_FRONT

    picture.depth = png_modes_to_bpp[mode]
    picture.data = png_byte_arr.getvalue()

    return picture


def create_request_session(max_retries=5) -> requests.Session:
    """Creates a request session that supports retry mechanism

    Args:
        max_retries (int, optional): Maximum retry times. Defaults to 5.

    Returns:
        requests.Session: Request session
    """
    session = requests.Session()
    adapter = HTTPAdapter(max_retries=max_retries)
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session



from io import BytesIO
import logging
from pathlib import Path
from typing import List, Optional

from PIL.Image import Image
from mutagen.flac import FLAC
from mutagen.id3 import APIC, ID3, ID3NoHeaderError, TALB, TDRC, TPE1, TPE2, TPOS, TRCK, TCON
from natsort import os_sorted

from dvtag.utils import (
    get_audio_paths_list,
    get_image,
    get_picture,
    get_png_byte_arr,
    get_rjid,
)

from .doujinvoice import DoujinVoice


def tag_mp3s(mp3_paths: List[Path], dv: DoujinVoice, png_bytes_arr: BytesIO, disc_number: Optional[int]):
    tags = ID3()

    tags.add(TALB(text=[dv.work_name]))
    tags.add(TPE2(text=[dv.circle]))
    tags.add(TDRC(text=[dv.sale_date]))
    if dv.genres:
        tags.add(TCON(text=[";".join(dv.genres)]))
    if disc_number:
        tags.add(TPOS(text=[str(disc_number)]))
    if dv.seiyus:
        tags.add(TPE1(text=dv.seiyus))
    tags.add(APIC(mime="image/png", desc="Front Cover", data=png_bytes_arr.getvalue()))

    for trck, p in enumerate(os_sorted(mp3_paths), start=1):
        tags.add(TRCK(text=[str(trck)]))

        try:
            if ID3(p) != tags:
                tags.save(p, v1=0)
                logging.info(f"Tagged <track: {trck}, disc: {disc_number}> to '{p.name}'")
        except ID3NoHeaderError:
            tags.save(p, v1=0)
            logging.info(f"Tagged <track: {trck}, disc: {disc_number}> to '{p.name}'")
        tags.delall("TRCK")


def tag_flacs(files: List[Path], dv: DoujinVoice, image: Image, png_bytes_arr: BytesIO, disc: Optional[int]):
    picture = get_picture(png_bytes_arr, image.width, image.height, image.mode)

    for trck, file in enumerate(os_sorted(files), start=1):
        tags = FLAC(file)
        tags.clear()
        tags.clear_pictures()

        tags.add_picture(picture)
        tags["album"] = [dv.work_name]
        tags["tracknumber"] = [str(trck)]
        tags["artist"] = dv.seiyus
        tags["albumartist"] = [dv.circle]
        tags["date"] = [dv.sale_date]
        tags["genre"] = dv.genres
        if disc:
            tags["discnumber"] = [str(disc)]

        if tags != FLAC(file):
            tags.save(file)
            logging.info(f"Tagged <track: {trck}, disc: {disc}> to '{file.name}'")


def tag(basepath: Path):
    rjid = get_rjid(basepath.name)
    dv = DoujinVoice(rjid)
    logging.info(f"[{rjid}] Ready to tag...")
    logging.info(f"Circle: {dv.circle}")
    logging.info(f"Album:  {dv.work_name}")
    logging.info(f"Seiyu:  {','.join(dv.seiyus)}")
    logging.info(f"Genre:  {','.join(dv.genres)}")
    logging.info(f"Date:   {dv.sale_date}")

    image = get_image(dv.work_image)
    png_bytes_arr = get_png_byte_arr(image)

    flac_paths_list, mp3_paths_list = get_audio_paths_list(basepath)

    set_disc = False
    disc = None
    if len(flac_paths_list) + len(mp3_paths_list) > 1:
        set_disc = True
        disc = 1

    for flac_files in flac_paths_list:
        tag_flacs(flac_files, dv, image, png_bytes_arr, disc)
        if set_disc:
            disc += 1

    for mp3_files in mp3_paths_list:
        tag_mp3s(mp3_files, dv, png_bytes_arr, disc)
        if set_disc:
            disc += 1

    logging.info(f"[{rjid}] Done.")
